
/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/scope.h>
#include <bld/block.h>
#include <bld/interp.h>

/***********************************************************************
class BLD_Block
***********************************************************************/
BLD_Block::BLD_Block(void)
	: terminatePrematurely(FALSE), index(0), prefix(0), suffix(0), inFail(FALSE)
{
}

GPL_Boolean BLD_Block::HasPrimarySubBlock(void) const
{
	if ((prefix + suffix) < subBlocks.GetCount())
	{
		return TRUE;
	}
	return FALSE;
}

void BLD_Block::AddRule(const BLD_RuleBase * rule,
						const GPL_Array<GPL_String> & expressions)
{
	// Build the new sub-block.
	BLD_SubBlock subBlock(rule->GetOperations(), expressions);

	switch(rule->GetType())
	{
	case BLD_RULE_FAIL_OPERATIONS:
		failBlocks.Append(subBlock);
		break;

	case BLD_RULE_PREPEND_OPERATIONS:
		subBlocks.Insert(subBlock, 0);
		++prefix;
		break;

	case BLD_RULE_APPEND_OPERATIONS:
		subBlocks.Append(subBlock);
		++suffix;
		break;

	default:
		// If there is an OVERWRITE sub-block...
		if (prefix + suffix < subBlocks.GetCount())
		{
			subBlocks.SetElement(subBlock, prefix);
		}

		// If there is no OVERWRITE sub-block...
		else
		{
			subBlocks.Insert(subBlock, prefix);
		}
		break;
	}
}

void BLD_Block::IncludeOperations(const OpArray & ops)
{
	if(!inFail)
	{
		if (index < subBlocks.GetCount())
		{
			subBlocks[index].IncludeOperations(ops);
		}
	}
	else
	{
		if (index < failBlocks.GetCount())
		{
			failBlocks[index].IncludeOperations(ops);
		}
	}
}

void BLD_Block::Execute(const GPL_Boolean reuseScope)
{
	// Premature termination is a game-time problem.
	terminatePrematurely = FALSE;

	index = 0;
	while (!terminatePrematurely && index < subBlocks.GetCount())
	{
		BLD_Scope scope;
		if (!reuseScope)
		{
			// Push the anonymous scope onto this context's stack.
			BLD_CONTEXT.Push(scope);
		}

		// Set the digit variables in the current scope.
		subBlocks[index].SetDigitVariables();

		// Perform each operation in the sub-block.
		subBlocks[index].Reset();
		while (subBlocks[index].Next())
		{
			// If externally terminated...
			if (terminatePrematurely)
			{
				break;
			}
		}

		if (!reuseScope)
		{
			// Pop the anonymous scope from this context's stack.
			BLD_CONTEXT.Pop(scope);
		}

		++index;
	}

	if(terminatePrematurely)
	{
		ExecuteFailBlocks(reuseScope);
	}
}

void BLD_Block::ExecuteFailBlocks(const GPL_Boolean reuseScope)
{
	inFail = TRUE;
	// Premature termination is a game-time problem.
	terminatePrematurely = FALSE;

	index = 0;
	while (!terminatePrematurely && index < failBlocks.GetCount())
	{
		BLD_Scope scope;
		if (!reuseScope)
		{
			// Push the anonymous scope onto this context's stack.
			BLD_CONTEXT.Push(scope);
		}

		// Set the digit variables in the current scope.
		failBlocks[index].SetDigitVariables();

		// Perform each operation in the sub-block.
		failBlocks[index].Reset();
		while (failBlocks[index].Next())
		{
			// If externally terminated...
			if (terminatePrematurely)
			{
				break;
			}
		}

		if (!reuseScope)
		{
			// Pop the anonymous scope from this context's stack.
			BLD_CONTEXT.Pop(scope);
		}

		++index;
	}
	inFail = FALSE;
}

/***********************************************************************
class BLD_SubBlock
***********************************************************************/
void BLD_SubBlock::SetDigitVariables(void)
{
	// A null scope name equates to the current scope,
	// whatever its name may be.
	GPL_String currentScope;

	GPL_String name;
	register const unsigned long n = expressions.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		name.Printf("%d", i);
		BLD_CONTEXT.SetValue(currentScope, name, BLD_Value(expressions[i]));
	}
}

void BLD_SubBlock::Reset(void)
{
	index = 0;
	shouldIncrement = TRUE;
}

GPL_Boolean BLD_SubBlock::Next(void)
{
	// If there are no more operations...
	if (index >= operations.GetCount())
	{
		return FALSE;
	}

	if (WBA_ARGS.Find("BLD_VERBOSE")->IsTrue())
	{
		WBA_APP->Printf("%s", operations[index]->ToString().GetData());
	}

	// Perform the next operation.
	operations[index]->Perform();

	if (!shouldIncrement)
	{
		shouldIncrement = TRUE;
	}
	else
	{
		++index;
	}

	return TRUE;
}

void BLD_SubBlock::IncludeOperations(const OpArray & ops)
{
	register const unsigned long n = operations.GetCount();
	if (index < n)
	{
		shouldIncrement = FALSE;
		operations.RemoveElement(index);
		operations.InsertRange(ops, index);
	}
}

void BLD_SubBlock::Dump(void) const
{
	for (unsigned long i = 0; operations.GetCount() > i; ++i)
	{
		WBA_APP->Printf("%d: %s", i, operations[i]->ToString().GetData());
	}
}
